/* engine/engine_hsm_pkey.c */

#include <libpki/pki.h>

/* Internal usage only - we want to keep the lib abstract */
#ifndef _LIBPKI_HSM_ENGINE_PKEY_H
#define _LIBPKI_HSM_ENGINE_PKEY_H

#define PKI_RSA_KEY	RSA
#define PKI_DSA_KEY	DSA

#ifdef ENABLE_ECDSA
#define PKI_EC_KEY	EC_KEY
#endif

#define PKI_RSA_KEY_MIN_SIZE		512
#define PKI_DSA_KEY_MIN_SIZE		512
#define PKI_EC_KEY_MIN_SIZE		56
PKI_DSA_KEY * _engine_pki_dsakey_new( int bits, ENGINE *e );
#ifdef ENABLE_ECDSA
PKI_EC_KEY * _engine_pki_ecdsakey_new( int bits, ENGINE *e);
#else
void * _engine_pki_ecdsakey_new( int bits, ENGINE *e );
#endif /* ENDIF::ENABLE_ECDSA */

int _engine_pki_rand_init( void );

#endif /* ENDIF::_LIBPKI_HSM_ENGINE_PKEY */

int _engine_pki_rand_seed( void ) {
	unsigned char seed[20];

	if (!RAND_pseudo_bytes(seed, 20)) {
		return 0;
	}
	RAND_seed(seed, sizeof seed);

	return(1);
}

PKI_RSA_KEY * _engine_pki_rsakey_new( int bits, ENGINE *e ) {
	BIGNUM *bn = NULL;
	unsigned long esp = 0x10001;
	PKI_RSA_KEY *rsa = NULL;

	if( bits < PKI_RSA_KEY_MIN_SIZE )
		return NULL;

	/*
	if( (bn = BN_new()) == NULL ) {
		// fprintf(stderr, "pollo\n");
		return NULL;
	};
	*/

	/*
	if( ( rsa = RSA_new()) == NULL ) {
		BN_free ( bn );
		// fprintf(stderr, "pollo2\n");
		return NULL;
	};
	*/

	/*
	if(!BN_set_word(bn,0x10001)) {
		BN_free( bn );
		// fprintf( stderr, "pollo3\n");
		return NULL;
	}
	*/

	// if( (rsa = RSA_generate_key(bits, bn, NULL, NULL)) == NULL ) {
	if( (rsa = RSA_generate_key(bits, esp, NULL, NULL)) == NULL ) {
		/* Error */
		// RSA_free( rsa );
		BN_free( bn );
		// fprintf(stderr, "pollo4\n");
		return NULL;
	}

	/* Free memory */
	// if( bn ) BN_free (bn);

	/* Let's return the RSA_KEY infrastructure */
	return (rsa);
}

PKI_DSA_KEY * _engine_pki_dsakey_new( int bits, ENGINE *e ) {
	PKI_DSA_KEY *k = NULL;
	unsigned char seed[20];

	if(bits < PKI_DSA_KEY_MIN_SIZE )  return NULL;

	if (!RAND_pseudo_bytes(seed, 20)) {
		/* Not enought rand ? */
		return NULL;
	}
	/*
	if((k = DSA_new()) == NULL ) {
		fprintf(stderr, "pla1\n");
		return NULL;
	}
	*/

	// if(!DSA_generate_parameters_ex( k, bits,
	if((k = DSA_generate_parameters( bits,
				seed, 20, NULL, NULL, NULL, NULL)) == NULL ) {
		if( k ) DSA_free( k );
		return NULL;
	}

	return( k );
}

#ifdef ENABLE_ECDSA
PKI_EC_KEY * _engine_pki_ecdsakey_new( int bits, ENGINE *e ) {
	/* ECDSA is a little more complicated than the other
	   schemes as it involves a group of functions. As the
	   purpose of this library is to provide a very hi-level
	   easy to use library, we will provide some hardwired
	   parameters.
	*/
	PKI_EC_KEY *k = NULL;
	EC_builtin_curve *curves = NULL;
	EC_GROUP *group = NULL;
	int num_curves = 0;
	int i = 0;
	unsigned int rnd_num = 0;
	int nid = 0;
	int degree = 0;
	int n_start = 0;

	if( bits < PKI_EC_KEY_MIN_SIZE ) return NULL;

	/* Get the number of availabe ECDSA curves in OpenSSL */
        if((num_curves = EC_get_builtin_curves(NULL, 0)) < 1 ) {
		/* No curves available! */
		return NULL;
	}

	/* Alloc the needed memory */
        curves = OPENSSL_malloc(sizeof(EC_builtin_curve) * num_curves);

	/* Check for memory allocation */
	if (curves == NULL) return NULL;

	/* Get the builtin curves */
	if (!EC_get_builtin_curves(curves, num_curves)) {
		if( curves ) free (curves);
		return NULL;
	}

	/* Get a random curve among the available ones */
	do {
		RAND_pseudo_bytes( (unsigned char *) &rnd_num, sizeof(rnd_num));
		i = rnd_num % num_curves;
		nid = curves[i].nid;
	} while ( nid == NID_ipsec4 );

	n_start = i;
	do {
		/* We have to match the strenght required for the
		   private key, let' cycle until we have curves */

		nid = curves[i].nid;

		/* Initialize the key */
		if ((k = EC_KEY_new()) == NULL) {
			if( curves) free (curves);
			return NULL;
		}

		if((group = EC_GROUP_new_by_curve_name(nid)) == NULL ) {
			if ( k ) EC_KEY_free( k );
			if ( curves ) free (curves);
			return NULL;
		}
		/* Assign the group to the key */
		if (EC_KEY_set_group(k, group) == 0) {
			if ( group ) EC_GROUP_free( group );
			if ( k ) EC_KEY_free( k );
			if ( curves ) free (curves);
			return NULL;
		}

		/* We do not need it now, let's free the group */
		if ( group ) EC_GROUP_free( group );
		group = NULL;

		degree = EC_GROUP_get_degree(EC_KEY_get0_group(k));

		if( degree < bits ) {
			/* drop the curve */
			EC_KEY_free(k);
			k = NULL;
		}

		/* Let's cycle through all the available curves
		   until we find one that matches (if any) */
		i = (i + 1 ) % num_curves;

	} while ( (degree < bits ) && (i != n_start) );

	/*
	fprintf(stderr, "%s:%d new i=%d [degree=%d]\n", 
			__FILE__, __LINE__, i, degree );
	*/

	if( degree < bits ) {
		/* No suitable curve found */
		if( curves ) free (curves );
		return NULL;
	}

	/* Now generate the key */
	if (!EC_KEY_generate_key(k)) {
		if( curves ) free (curves);
		if( k ) EC_KEY_free (k);
		PKI_log_debug("ERROR, generating ecdsa key\n");
		return NULL;
	}

	/* Verify the Key to be ok */
	if (!EC_KEY_check_key(k)) {
		if( curves ) free (curves);
		if( k ) EC_KEY_free (k);
		PKI_log_debug("ERROR, verifying ecdsa key\n");
		return NULL;
	}

	/*
	if(( k = EC_KEY_new_by_curve_name(i)) == NULL ) {
		fprintf(stderr, "olo1\n");
		return NULL;
	}
	*/

	return ( k );
}

#else /* EVP_PKEY_EC */

void * _engine_pki_ecdsakey_new( int bits, ENGINE *e ) {
	return ( NULL );
}

#endif


PKI_X509_KEYPAIR *HSM_ENGINE_X509_KEYPAIR_new( int type, int bits, 
				URL *url, PKI_CRED *cred, HSM *driver ) {

	PKI_X509_KEYPAIR *ret = NULL;
	PKI_X509_KEYPAIR_VALUE *val = NULL;
	PKI_RSA_KEY *rsa = NULL;
	PKI_DSA_KEY *dsa = NULL;
#ifdef ENABLE_ECDSA
	PKI_EC_KEY *ec = NULL;
#endif
	ENGINE *e = NULL;

	if((val = EVP_PKEY_new()) == NULL ) {
		return NULL;
	}

	e = (ENGINE *) driver;
	if( _engine_pki_rand_seed() == 0 ) {
		/* Probably low level of randomization available */
		PKI_log_debug("WARNING, low rand available (ENGINE HSM)");
	}

	switch (type) {
		case PKI_SCHEME_RSA:
			if((rsa = _engine_pki_rsakey_new( bits, e )) == NULL ) {
				if( val ) EVP_PKEY_free( val );
				return NULL;
			};
			if(!EVP_PKEY_assign_RSA( (EVP_PKEY *) val, rsa)) {
				if( rsa ) RSA_free( rsa );
				if( val ) EVP_PKEY_free( val );
				return NULL;
			};
			break;
		case PKI_SCHEME_DSA:
			if((dsa = _engine_pki_dsakey_new( bits, e )) == NULL ) {
				if( val ) EVP_PKEY_free( val );
				return(NULL);
			};
			if (!DSA_generate_key( dsa )) {
				if( val ) EVP_PKEY_free( val );
				return NULL;
			}
                        if (!EVP_PKEY_assign_DSA( (EVP_PKEY *) val, dsa)) {
				if( dsa ) DSA_free ( dsa );
				if( val ) EVP_PKEY_free( val );
				return NULL;
			}
                        dsa=NULL;
			break;
#ifdef ENABLE_ECDSA
		case PKI_SCHEME_ECDSA:
			if((ec = _engine_pki_ecdsakey_new(bits, e)) == NULL ) {
				if( val ) EVP_PKEY_free( val );
				return(NULL);
			};
			if (!EC_KEY_generate_key(ec)) {
				if( val ) EVP_PKEY_free( val );
				return NULL;
			}
                        if (!EVP_PKEY_assign_EC_KEY( (EVP_PKEY *) val, ec)) {
				if( ec ) EC_KEY_free ( ec );
				if( val ) EVP_PKEY_free( val );
				return NULL;
			}
			ec=NULL;
			break;
#endif
		default:
			/* No recognized scheme */
			if( val ) EVP_PKEY_free( val );
			return NULL;
	}

	if((ret = PKI_X509_new( PKI_DATATYPE_X509_KEYPAIR, driver)) == NULL ) {
		if( val ) EVP_PKEY_free ( val );
		return NULL;
	}

	ret->value = val;

	/* Let's return the PKEY infrastructure */
	return ( ret );
}

/* Key Free function */
void HSM_ENGINE_X509_KEYPAIR_free ( PKI_X509_KEYPAIR *pkey ) {

	if( !pkey) return;

	PKI_X509_KEYPAIR_free( pkey);

	return;

}

