/* PKI_X509 object management */

#include <libpki/pki.h>
#include <sys/utsname.h>

/*! \brief Generate a new CRL from a stack of revoked entries
 *
 * Generates a new signed CRL from a stack of revoked entries. A profile is
 * used to set the right extensions in the CRL. To generate a new revoked
 * entry the PKI_X509_CRL_ENTRY_new() function has to be used.
 */

PKI_X509_CRL *PKI_X509_CRL_new ( PKI_X509_KEYPAIR *k, PKI_X509_CERT *cert, 
	char * serial_s, unsigned long validity, PKI_X509_CRL_ENTRY_STACK *sk, 
			PKI_X509_PROFILE *profile, PKI_CONFIG *oids, HSM *hsm) {

	PKI_X509_CRL *ret = NULL;
	PKI_X509_CRL_VALUE *val = NULL;
	ASN1_INTEGER *serial = NULL;
	ASN1_TIME *time = NULL;
	PKI_X509_KEYPAIR_VALUE *pkey = NULL;
	int rv = PKI_OK;

	PKI_DIGEST_ALG *dgst = NULL;

	long long total_validity = 0;

	/* Checks for the Key and its internal value */
	if( !k || !k->value ) return NULL;
	pkey = k->value;

	/* checks for the certificate and its internal value */
	if( !cert || !cert->value ) return ( NULL );

	/* Alloc memory structure for the Certificate */
        val = (PKI_X509_CRL_VALUE *) X509_CRL_new();
        if( !val ) {
                PKI_log_debug("ERROR, memory allocation!");
                return (NULL);
        }

	if( serial_s ) {
                serial = s2i_ASN1_INTEGER( NULL, serial_s);
        } else {
                serial = s2i_ASN1_INTEGER( NULL, "0");
        }

	/* Set the start date NOW */
	if((time = ASN1_TIME_new()) == NULL ) {
		PKI_log_debug ( "ERROR, memory allocation!");
		return ( NULL );
	}

	/* Get the current time */
        if(X509_gmtime_adj( time, 0) == NULL ) {
                PKI_log_debug("ERROR, can not set current time!");
		PKI_X509_CRL_free ( ret );
		ASN1_TIME_free ( time );
        }

	/* Set the Last Update field */
	if(X509_CRL_set_lastUpdate( val, time ) == 0 ) {
		PKI_log_err ( "ERROR, can not set lastUpdate field in CRL");
		PKI_X509_CRL_free ( ret );
		ASN1_TIME_free ( time );
		return ( NULL );
	}

	if( (validity > 0) || (profile == NULL) ) {
		total_validity = validity;
	} else {
		char *years_s = NULL;
		char *days_s = NULL;
		char *hours_s = NULL;
		char *mins_s = NULL;
		char *secs_s = NULL;

		long long years = 0;
		long long days = 0;
		long long hours = 0;
		long long mins = 0;
		long long secs = 0;

		if((years_s = PKI_CONFIG_get_attribute_value( profile, 
					"validity", "years" )) != NULL ) {
			years = atoll( years_s );
		}

		if((days_s = PKI_CONFIG_get_attribute_value( profile, 
					"validity", "days" )) != NULL ) {
			days = atoll( days_s );
		}

		if((hours_s = PKI_CONFIG_get_attribute_value( profile, 
					"validity", "hours" )) != NULL ) {
			hours = atoll( hours_s );
		}

		if((mins_s = PKI_CONFIG_get_attribute_value( profile, 
					"validity", "mins" )) != NULL ) {
			mins = atoll( mins_s );
		}

		if((secs_s = PKI_CONFIG_get_attribute_value( profile, 
					"validity", "secs" )) != NULL ) {
			secs = atoll( secs_s );
		}

		total_validity = secs + 
				60 * ( mins + 
					60 * (hours + 
						24 * ( days + 
							365 * years 
						     )
					     )
				    );

		if( years_s ) PKI_Free ( years_s );
		if( hours_s ) PKI_Free ( hours_s );
		if( mins_s ) PKI_Free ( mins_s );
		if( secs_s ) PKI_Free ( secs_s );
	};

	if( X509_gmtime_adj ( time, (long int) total_validity ) == NULL ) {
		PKI_log_debug ( "ERROR, can not set gmtime for nextUpdate");
		X509_CRL_free ( val );
		ASN1_TIME_free ( time );
		return ( NULL );
	}

	if( X509_CRL_set_nextUpdate( val, time ) == 0 ) {
		PKI_log_err( "Can not set nextUpdate field in CRL");
		X509_CRL_free ( val );
		ASN1_TIME_free ( time );
		return ( NULL );
	}

	/* We do not need it anymore */
	ASN1_TIME_free ( time );

	/* Now we need to add the CRL issuer name and details */
	if (X509_CRL_set_issuer_name( val, 
				X509_get_subject_name(cert->value)) == 0) {
		PKI_log_debug( "Can not set CRL issuer name");
		X509_CRL_free ( val );
		return ( NULL );
	}

	if((ret = PKI_X509_new( PKI_DATATYPE_X509_CRL, hsm)) == NULL ) {
		PKI_log_err ("Memory Error");
		X509_CRL_free ( val );
		return ( NULL );
	}

	ret->value = val;

	/* Get the extensions from the profile */
	if( profile ) {
		PKI_TOKEN * tk;

		if((tk = PKI_TOKEN_new_null()) == NULL ) {
			PKI_log_err ( "Memory allocation failure");
			PKI_X509_CRL_free ( ret );
			return NULL;
		}

		PKI_TOKEN_set_cert ( tk, cert );
		PKI_TOKEN_set_keypair ( tk, k );

		if(PKI_X509_EXTENSIONS_crl_add_profile( profile, oids, ret, tk) == 0 ) {
					PKI_log_debug( "ERROR, can not set extensions!");
			PKI_X509_CRL_free ( ret );

			tk->cert = NULL;
			tk->keypair = NULL;
			PKI_TOKEN_free ( tk );

			return ( NULL );
		}
		tk->cert = NULL;
		tk->keypair = NULL;
		PKI_TOKEN_free ( tk );
	}

	/* Get the Digest Algorithm */
	if( (dgst = PKI_DIGEST_ALG_get_by_key( k )) == NULL ) {
		PKI_log_err("Can not get digest algor from keypair!");
		PKI_X509_CRL_free ( ret ); 
		return ( NULL );
	}
	
	rv = PKI_X509_sign ( ret, dgst, k );

	/*
	rv = PKI_sign ( PKI_DATATYPE_X509_CRL, val->crl, 
			(void *) &X509_CRL_INFO_it, val->crl->sig_alg,
				val->sig_alg, val->signature, k, dgst);
	*/

	if ( rv == PKI_ERR ) {
		PKI_log_debug ("ERROR, can not sign CRL!");
		PKI_X509_CRL_free ( ret );
		return NULL;
	}

	return( ret );
}


/*! \brief Generates a new PKI_X509_CRL_ENTRY from a certificate
 *
 * This function generates a new PKI_X509_CRL_ENTRY starting from a certificate.
 * The new structure can be added to a PKI_X509_CRL_ENTRY_STACK structure in
 * order to generate a new CRL.
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_ENTRY_new ( PKI_X509_CERT *cert, 
		PKI_X509_CRL_REASON reason, PKI_TIME * revDate,
						PKI_X509_PROFILE *profile ) {

	PKI_X509_CRL_ENTRY *entry = NULL;
	char *ser_s = NULL;

	if((ser_s = PKI_X509_CERT_get_parsed(cert, PKI_X509_DATA_SERIAL)) 
								== NULL ) {
		return ( NULL );
	}

	entry = PKI_X509_CRL_ENTRY_new_serial ( ser_s, reason, 
							revDate, profile );

	if( ser_s ) PKI_Free ( ser_s );

	return ( entry );
}

/*! \brief Generates a new PKI_X509_CRL_ENTRY from a serial number (string)
 *
 * This function generates a new PKI_X509_CRL_ENTRY starting from a string
 * representing the serial number of the entry. The optional profile
 * is used to add extensions to the entry (when needed)
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_ENTRY_new_serial ( char *serial, 
		PKI_X509_CRL_REASON reason, PKI_TIME *revDate,
					PKI_X509_PROFILE *profile ) {

	PKI_X509_CRL_ENTRY *entry = NULL;

	if( !serial ) return (NULL);

	if((entry = (PKI_X509_CRL_ENTRY *) X509_REVOKED_new()) == NULL ) {
		/* Memory alloc error */
		return (NULL);
	}

	if((entry->serialNumber = s2i_ASN1_INTEGER( NULL, serial )) == NULL ) {
		goto err;
	}

	if( revDate ) {
		if( reason == CRL_REASON_HOLD_INSTRUCTION ) {
                	if (!X509_REVOKED_add1_ext_i2d ( entry, 
					NID_invalidity_date, revDate, 0, 0)) {
				goto err;
			}
			if (!X509_REVOKED_add1_ext_i2d(entry, 
				NID_hold_instruction_code, 
				PKI_OID_get( "holdInstructionReject"), 0, 0)) {
                        		goto err;
                	}
		} else {
			if(!X509_REVOKED_set_revocationDate( 
					(X509_REVOKED *) entry, 
						(ASN1_TIME *) revDate)) {
				goto err;
			}
		}
	}

/*
	if (rev && !X509_REVOKED_set_revocationDate(rev, revDate))
                goto err;

        if (rev && (reason_code != OCSP_REVOKED_STATUS_NOSTATUS))
                {
                rtmp = ASN1_ENUMERATED_new();
                if (!rtmp || !ASN1_ENUMERATED_set(rtmp, reason_code))
                        goto err;
                if (!X509_REVOKED_add1_ext_i2d(rev, NID_crl_reason, rtmp, 0, 0))
                        goto err;
                }

        if (rev && comp_time)
                {
                if (!X509_REVOKED_add1_ext_i2d(rev, NID_invalidity_date, comp_time, 0, 0))
                        goto err;
                }
	if (rev && hold)
                {
                if (!X509_REVOKED_add1_ext_i2d(rev, NID_hold_instruction_code, hold, 0, 0))
                        goto err;
                }

*/

	return( entry );

err:

	if( entry ) X509_REVOKED_free ( (X509_REVOKED *) entry );
	return (NULL);
}

void PKI_X509_CRL_ENTRY_free_void ( void *entry ) {
	PKI_X509_CRL_ENTRY_free ( (PKI_X509_CRL_ENTRY *) entry );
}

/*! \brief Frees a PKI_X509_CRL_ENTRY
 *
 * Frees memory associated to a PKI_X509_CRL_ENTRY
 */

int PKI_X509_CRL_ENTRY_free ( PKI_X509_CRL_ENTRY *entry ) {

	if( !entry ) return (PKI_ERR);

	if( entry ) X509_REVOKED_free ( (X509_REVOKED *) entry );

	return (PKI_OK);

}

/*! \brief Find an entry within a CRL by using the PKI_INTEGER serial
 *         number of the certificate
 *
 * This function look inside a CRL and returns the PKI_X509_CRL_ENTRY
 * if the passed serial is found.  The returned pointer refers to the
 * internal CRL structure, when the CRL is freed the pointer will no
 * more point to a valid memory area.
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_lookup( PKI_X509_CRL *x, 
						PKI_INTEGER *s ) {

	PKI_X509_CRL_ENTRY *r = NULL;
	X509_CRL *crl = NULL;

        long long curr = 0;
        long long end, cmp_val;

	if( !x || !s ) return (NULL);

	/* Set the end point to the last one */
        end = (long long) sk_X509_REVOKED_num(crl->crl->revoked) - 1;
        if( end < 0 ) return (r);

	crl = (X509_CRL *) x;

        /* Look for serial number of certificate in CRL */
        // rtmp.serialNumber = (ASN1_INTEGER *) serial;
        // ok = sk_X509_REVOKED_find(crl->crl->revoked, &rtmp);

	for( curr = 0 ; curr <= end ; curr++ ) {
		r = sk_X509_REVOKED_value( crl->crl->revoked, (int) curr );
                if(( cmp_val = ASN1_INTEGER_cmp(r->serialNumber, s)) == 0 ) {
			/* Found */
			break;
		}
	}

	return( r );
}

/*! \brief Find an entry within a CRL
 *
 * This function look inside a CRL and returns the PKI_X509_CRL_ENTRY
 * if the passed serial is found.  The returned pointer refers to the
 * internal CRL structure, when the CRL is freed the pointer will no
 * more point to a valid memory area.
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_lookup_serial( PKI_X509_CRL *x, 
							char *serial ) {

	ASN1_INTEGER *s = NULL;
	PKI_X509_CRL_ENTRY *r = NULL;

	if( !x || !serial ) return (NULL);

	if((s = PKI_INTEGER_new_char ( serial )) == NULL ) {
		return ( NULL );
	}

	r = PKI_X509_CRL_lookup ( x, s );

	if ( s ) PKI_INTEGER_free ( s );

	return( r );
}

/*! \brief Find an entry in a CRL by using a certificate
 *
 * Use the information within the certificate to look for a corresponding
 * entry in the CRL. If found, the PKI_X509_CRL_ENTRY structure is copied and
 * returned.  The returned pointer refers to the internal CRL structure, 
 * when the CRL is freed the pointer will no more point to a valid memory
 * area.
 */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_lookup_cert ( PKI_X509_CRL *x, 
							PKI_X509_CERT *cert ) {

	// ASN1_INTEGER *serial = NULL;
	PKI_INTEGER *serial = NULL;

	if( !x || !cert ) return (NULL);

	if((serial = PKI_X509_CERT_get_data ( cert, 
			PKI_X509_DATA_SERIAL )) == NULL ) {
		return ( NULL );
	}

	/*
	if((serial = X509_get_serialNumber( (X509 *) cert )) == NULL ) {
		return (NULL);
	}

	if((serial_s = i2s_ASN1_INTEGER( NULL, serial )) == NULL ) {
		return(NULL);
	}

	entry = PKI_X509_CRL_lookup_serial( x, serial_s );
	*/

	/* Now let's free the serial_s */
	// if( serial_s ) PKI_Free ( serial_s );

	return PKI_X509_CRL_lookup ( x, serial );
	// return( entry );
}

/*! \brief Lookup for a serial (long long) in a PKI_X509_CRL and returns the
 *         PKI_X509_CRL_ENTRY entry if found */

PKI_X509_CRL_ENTRY * PKI_X509_CRL_lookup_long(PKI_X509_CRL *x, long long s ) {

	PKI_INTEGER * serial = NULL;
	PKI_X509_CRL_ENTRY *entry = NULL;

	if ( !x ) return NULL;

	serial = PKI_INTEGER_new ( s );

	entry = PKI_X509_CRL_lookup ( x, serial );

	if ( serial ) PKI_INTEGER_free ( serial );

	return ( entry );
}

void PKI_X509_CRL_free_void ( void * x ) {
	PKI_X509_CRL_free ( (PKI_X509_CRL *) x );
}

/*! \brief Frees memory associated with a PKI_CRL
 *
 * This function frees memory associated with a PKI_X509_CRL data structure.
 * Returns PKI_OK if successful or PKI_ERR in case of an error (or if
 * the passed pointer was NULL.
 */

int PKI_X509_CRL_free ( PKI_X509_CRL * x ) {

	if( !x ) return (PKI_ERR);

	if( x ) X509_CRL_free ((X509_CRL *) x );

	return( PKI_OK );
}

/*! \brief Adds an Extension to a CRL object
 */

int PKI_X509_CRL_add_extension(PKI_X509_CRL *x, PKI_X509_EXTENSION *ext) {

	if( !x || !x->value || !ext || !ext->value ) return (PKI_ERR);

	if (!X509_CRL_add_ext((X509_CRL *)x->value, ext->value, -1)) 
		return (PKI_ERR);

	return (PKI_OK);
}

/*! \brief Adds a stack of extensions to a CRL object.
 */

int PKI_X509_CRL_add_extension_stack(PKI_X509_CRL *x, 
					PKI_X509_EXTENSION_STACK *ext) {

	int i = 0;

	if( !x || !ext ) return (PKI_ERR);

	for( i = 0; i < PKI_STACK_X509_EXTENSION_elements(ext); i++ ) {
		PKI_X509_EXTENSION *ossl_ext = NULL;

		ossl_ext = PKI_STACK_X509_EXTENSION_get_num( ext, i);
		if (!ossl_ext ) continue;

		if(!X509_CRL_add_ext ( (X509_CRL *) x->value, 
							ossl_ext->value, -1 )) {
			PKI_log_err("Adding Extension::%s",
				ERR_error_string(ERR_get_error(), NULL));
			return ( PKI_ERR );
		}

	}

	return (PKI_OK);
}

/*! \brief Get Data from a CRL object
 */

void * PKI_X509_CRL_get_data ( PKI_X509_CRL *x, PKI_X509_DATA type ) {

	void *ret = NULL;
	PKI_X509_CRL_VALUE *tmp_x = NULL;
	PKI_MEM *mem = NULL;

	if( !x || !x->value ) return (NULL);

	tmp_x = x->value;

	switch( type ) {
		case PKI_X509_DATA_VERSION:
			ret = (tmp_x)->crl->version;
			// ret = X509_CRL_get_version((X509_CRL *) x);
			break;
		case PKI_X509_DATA_ISSUER:
			ret = X509_CRL_get_issuer ((X509_CRL *) tmp_x);
			break;
		case PKI_X509_DATA_NOTAFTER:
		case PKI_X509_DATA_NEXTUPDATE:
			ret = X509_CRL_get_nextUpdate ((X509_CRL *) tmp_x);
			break;
		case PKI_X509_DATA_NOTBEFORE:
		case PKI_X509_DATA_LASTUPDATE:
			ret = X509_CRL_get_lastUpdate ((X509_CRL *) tmp_x);
			break;
		case PKI_X509_DATA_EXTENSIONS:
			// ret = X509_CRL_get_extensions((X509_CRL *) x );
			PKI_log_debug("To Be implemented!");
			return ( NULL );
			// ret = X509_CRL_get_algor((X509_CRL *) x );
			break;
		case PKI_X509_DATA_SIGNATURE:
			ret = tmp_x->signature;
			// ret = X509_CRL_get_signature ((X509_CRL *) x);
			break;
		case PKI_X509_DATA_SIGNATURE_ALG1:
			if ( tmp_x->crl ) {
				ret = tmp_x->crl->sig_alg;
			}
			break;
		case PKI_X509_DATA_ALGORITHM:
		case PKI_X509_DATA_SIGNATURE_ALG2:
			ret = tmp_x->sig_alg;
			break;
		case PKI_X509_DATA_TBS_MEM_ASN1:
			if((mem = PKI_MEM_new_null()) == NULL ) 
				break;
			mem->size = ASN1_item_i2d ( (void *) tmp_x->crl, 
				&(mem->data), &X509_CRL_INFO_it );
			ret = mem;
			break;
		default:
			/* Not Recognized/Supported DATATYPE */
			return (NULL);
	}

	return (ret);
}

char * PKI_X509_CRL_get_parsed( PKI_X509_CRL *x, PKI_X509_DATA type ) {

	char *ret = NULL;
	PKI_X509_CRL_VALUE *x_val = NULL;
	PKI_ALGOR *al = NULL;

	if( !x || !x->value ) return (NULL);

	x_val = x->value;

	switch( type ) {
		case PKI_X509_DATA_VERSION:
			ret = PKI_INTEGER_get_parsed ( PKI_X509_CRL_get_data(
					x, type ));
			if (!ret) ret = strdup ("NONE");
			break;
        	case PKI_X509_DATA_ISSUER:
			ret = PKI_X509_NAME_get_parsed( PKI_X509_CRL_get_data ( 
					x, type));
			if (!ret) ret = strdup ("NONE");
			break;
		case PKI_X509_DATA_ALGORITHM:
			if((al = PKI_X509_CRL_get_data ( x, type )) != NULL ) {
				ret = strdup ( PKI_OID_get_descr ( al->algorithm));
			} else {
				ret = strdup ("NONE");
			}
			break;
		case PKI_X509_DATA_LASTUPDATE:
		case PKI_X509_DATA_NEXTUPDATE:
		case PKI_X509_DATA_NOTBEFORE:
		case PKI_X509_DATA_NOTAFTER:
			ret = PKI_TIME_get_parsed( PKI_X509_CRL_get_data (
					x, type ));
			if (!ret) ret = strdup ("NONE");
			break;
		default:
			/* Not Recognized/Supported DATATYPE */
			return (NULL);
	}

	return (ret);
}

int PKI_X509_CRL_print_parsed(PKI_X509_CRL *x, PKI_X509_DATA type, 
								int fd){

	const char *str = NULL;
	int rv = 0;

	if ( !x ) return ( PKI_ERR );

	switch ( type ) {
		default:
			/* Not Recognized/Supported DATATYPE */
			return ( PKI_ERR );
	}

	if( str ) {
		rv = write( fd, str, strlen(str) );
	}

	return ( PKI_OK );
}

/*! \brief Puts a CRL in a PKI_MEM structure */
/*
PKI_MEM *PKI_X509_CRL_put_mem ( PKI_X509_CRL *x, int format ) {

	PKI_X509_CRL_STACK *sk = NULL;
	PKI_MEM *pki_mem = NULL;

	if( !x ) return ( NULL );

	if((sk = PKI_STACK_X509_CRL_new()) == NULL ) {
		PKI_log_err ( "PKI_X509_CRL_STACK_put_mem()::Memory Error!");
		return ( NULL );
	}

	if( PKI_STACK_X509_CRL_push ( sk, x ) == PKI_ERR ) {
		PKI_log_err ( "PKI_X509_CRL_STACK_put_mem()::Memory Error!");
		if( sk ) PKI_STACK_X509_CRL_free ( sk );
		return ( NULL );
	}

	if((pki_mem = PKI_MEM_new_null()) == NULL ) {
		PKI_log_err ( "PKI_X509_CRL_STACK_put_mem()::Memory Error!");
		if( sk ) PKI_STACK_X509_CRL_free ( sk );
		return ( NULL );
	}

	if((PKI_X509_CRL_STACK_put_mem(sk, format, pki_mem, NULL, 0)) 
								== PKI_ERR) {
		if( sk ) PKI_STACK_X509_CRL_free ( sk );
		if ( pki_mem ) PKI_MEM_free ( pki_mem );
		return ( NULL );
	}

	if ( sk ) PKI_STACK_X509_CRL_free_all (sk);

	return ( pki_mem );
}
*/
