/* TOKEN Object Management Functions */

#include <libpki/pki.h>

const long LIBPKI_OS_DETAILS = LIBPKI_OS_CLASS | 
		LIBPKI_OS_BITS | LIBPKI_OS_VENDOR;

#include <sys/types.h>
#include <dirent.h>

static int _libpki_init = 0;

#if OPENSSL_VERSION_NUMBER < 0x00908000
int NID_proxyCertInfo = -1;
#endif

static char *libpki_oids[] = {
        "1.3.6.1.4.1.311.20.2", "certificateTemplate", "Certificate Template",
	/* OID - {{1.3.6.1.4.1.18227 (OpenCA)} . 50 (Extensions)} */
	"1.3.6.1.4.1.18227.50.1", "loa", "Level of Assurance",
	"1.3.6.1.4.1.18227.50.2", "certificateUsage", "Certificate Usage",
	NULL, NULL, NULL
};

static int __init_add_libpki_oids ( void ) {
	int i, ret;

        i = 0;

        while( libpki_oids[i] && libpki_oids[i+1] && libpki_oids[i+2] ) {
        	if((ret = OBJ_create(libpki_oids[i], libpki_oids[i+1], 
					libpki_oids[i+2])) == NID_undef) {
        		return 0;
        	}
        	i = i+3;
	}

	/* Special Case for proxyCertInfo NID */
#if OPENSSL_VERSION_NUMBER < 0x00908000
	NID_proxyCertInfo = OBJ_create( "1.3.6.1.5.5.7.1.14", 
			"proxyCertInfo", "Proxy Certificate Information");
#endif
	return PKI_OK;
}

// static int _list_all_tokens_dir ( char * dir, PKI_STACK *list );

/*!
 * \brief Initialize libpki internal structures.
 */

int PKI_init_all( void ) {

	/* Initialize OpenSSL so that it adds all the needed algor and dgst */
	if( _libpki_init == 0 ) {
		X509V3_add_standard_extensions();
		OpenSSL_add_all_algorithms();
		OpenSSL_add_all_digests();
		OpenSSL_add_all_ciphers();
		OpenSSL_pthread_init();

		ERR_load_ERR_strings();
		ERR_load_crypto_strings();

		PRQP_init_all_services();
		PKI_X509_SCEP_init();
		xmlInitParser();

		__init_add_libpki_oids ();
	}

	/* Enable Proxy Certificates Support */
	PKI_set_env( "OPENSSL_ALLOW_PROXY", "1");

	/* Set the initialization bit */
	_libpki_init = 1;

	/* Check Application and LibPKI coherence */
	if( (LIBPKI_OS_CLASS | LIBPKI_OS_BITS | LIBPKI_OS_VENDOR ) !=
							LIBPKI_OS_DETAILS ) {
		PKI_log_err ("WARNING::LibPKI and Application OS details are "
				"different [%d/%d]", LIBPKI_OS_DETAILS,
				LIBPKI_OS_CLASS | LIBPKI_OS_BITS | 
					LIBPKI_OS_VENDOR);
	};

	return ( PKI_OK );
}

/*! \brief Returns the status of the library (check for initialization) */

int PKI_get_init_status ( void ) {

	if( _libpki_init == 0 ) {
		return PKI_STATUS_NOT_INIT;
	}

	return PKI_STATUS_INIT;
}

/* \brief Returns a stack of the names of all the available tokens */

PKI_STACK * PKI_list_all_tokens ( char *dir ) {

	char *name = NULL;

	PKI_STACK *dir_list = NULL;
	PKI_STACK *list = NULL;
	int i = 0;

	if((dir_list = PKI_CONFIG_get_search_paths( dir )) == NULL ) {
		return ( NULL );
	}

	if((list = PKI_STACK_new_null()) == NULL ) {
		return ( NULL );
	}

	for ( i=0; i <PKI_STACK_elements( dir_list ); i++ ) {
		if((name = PKI_STACK_get_num ( dir_list, i )) != NULL ) {
			PKI_list_all_tokens_dir ( name, list );
		}
		PKI_Free ( name );
	}

	if( dir_list ) PKI_STACK_free ( dir_list );

	return ( list );
}

PKI_STACK * PKI_list_all_tokens_dir ( char * dir, PKI_STACK *list ) {

        struct dirent *dd = NULL;
	DIR *dirp = NULL;
	URL *url = NULL;

	char *token_dir = NULL;
	size_t token_dir_size = 0;

	PKI_STACK *ret = NULL;

	if( !dir ) return ( NULL );

	if(( url = URL_new ( dir )) == NULL ) {
		return (NULL);
	}

	if( url->proto != URI_PROTO_FILE ) {
		if( url ) URL_free (url );
		return (NULL);
	}

	if( !list ) {
		if((ret = PKI_STACK_new_null()) == NULL ) {
			if( url ) URL_free (url );
			return( NULL );
		}
	} else {
		ret = list;
	}

	token_dir_size = strlen(url->addr) + 1 +
				strlen( PKI_DEFAULT_TOKEN_DIR ) + 1;

	token_dir = PKI_Malloc ( token_dir_size );
	snprintf( token_dir, token_dir_size, "%s/%s",
			url->addr, PKI_DEFAULT_TOKEN_DIR );

	PKI_log_debug("PKI_list_all_tokens_dir()::Opening dir %s", token_dir);

	if((dirp = opendir( token_dir )) == NULL ) {

		snprintf( token_dir, token_dir_size, "%s", url->addr );

		PKI_log_debug("PKI_list_all_tokens_dir()::Opening dir %s", token_dir);
		if((dirp = opendir( token_dir)) == NULL ) {
			if( url ) URL_free (url );
			if( token_dir ) PKI_Free ( token_dir );
			return ( ret );
		}
	}

	while(( dd = readdir( dirp )) != NULL ) {
		long len;
		char *filename = NULL;
		PKI_TOKEN *tk = NULL;

		filename = dd->d_name;
		len = (long) strlen( filename );

		if( (len < 4) || (strcmp( ".xml", filename +len -4 ))) {
			continue;
		} else {
			char fullpath[BUFF_MAX_SIZE];
			size_t fullsize = 0;

			PKI_CONFIG *tmp_cfg = NULL;
			char *tmp_name = NULL;

			snprintf(fullpath, BUFF_MAX_SIZE,
				"%s/%s", token_dir, filename );

			if((fullsize = strlen(token_dir) + 
				strlen( filename ) + 1) > 
						BUFF_MAX_SIZE) {
				continue;
			}
				
			if((tmp_cfg = PKI_CONFIG_load( fullpath )) ==
								NULL ) {
				continue;
			}

			if((tmp_name = PKI_CONFIG_get_value( tmp_cfg,
						"/*/name")) != NULL) {

				if((tk = PKI_TOKEN_new_null()) != NULL ) {
					if((PKI_TOKEN_init( tk, token_dir, tmp_name )) != PKI_ERR ) {
						PKI_STACK_push( list, strdup(tmp_name));
					}
					PKI_TOKEN_free( tk );
				}
			}

		}
	}
	closedir( dirp );

	if( url ) URL_free (url);
	if( token_dir ) PKI_Free ( token_dir );

	return ( ret );
}


/* \brief Returns a stack of all the available PKI_TOKENS */

PKI_TOKEN_STACK *PKI_get_all_tokens ( char *dir ) {
	char *name = NULL;

	PKI_STACK *dir_list = NULL;
	PKI_TOKEN_STACK *list = NULL;
	int i = 0;

	if((dir_list = PKI_CONFIG_get_search_paths( dir )) == NULL ) {
		return ( NULL );
	}

	if((list = PKI_STACK_TOKEN_new_null()) == NULL ) {
		return ( NULL );
	}

	for ( i=0; i <PKI_STACK_elements( dir_list ); i++ ) {
		if((name = PKI_STACK_get_num ( dir_list, i )) != NULL ) {
			PKI_get_all_tokens_dir ( name, list );
		}
		PKI_Free ( name );
	}

	if( dir_list ) PKI_STACK_free ( dir_list );

	return ( list );
}

PKI_TOKEN_STACK *PKI_get_all_tokens_dir ( char *dir, PKI_TOKEN_STACK *list ) {

	struct dirent *dd = NULL;
	DIR *dirp = NULL;
	URL *url = NULL;

	char *token_dir = NULL;
	size_t token_dir_size = 0;

	PKI_TOKEN_STACK *ret = NULL;

	if( !dir ) return ( NULL );

	if(( url = URL_new ( dir )) == NULL ) {
		return (NULL);
	}

	if( url->proto != URI_PROTO_FILE ) {
		if( url ) URL_free (url );
		return (NULL);
	}

	if( !list ) {
		if((ret = PKI_STACK_TOKEN_new_null()) == NULL ) {
			if( url ) URL_free (url );
			return( NULL );
		}
	} else {
		ret = list;
	}

	token_dir_size = strlen(url->addr) + 1 +
				strlen( PKI_DEFAULT_TOKEN_DIR ) + 1;

	token_dir = PKI_Malloc ( token_dir_size );
	snprintf( token_dir, token_dir_size, "%s/%s",
			url->addr, PKI_DEFAULT_TOKEN_DIR );

	PKI_log_debug("PKI_list_all_tokens_dir()::Opening dir %s", token_dir);

	if((dirp = opendir( token_dir )) == NULL ) {

		snprintf( token_dir, token_dir_size, "%s", url->addr );

		PKI_log_debug("PKI_list_all_tokens_dir()::Opening dir %s", token_dir);
		if((dirp = opendir( token_dir)) == NULL ) {
			if( url ) URL_free (url );
			if( token_dir ) PKI_Free ( token_dir );
			return ( ret );
		}
	}

	while(( dd = readdir( dirp )) != NULL ) {
		long len;
		char *filename = NULL;
		PKI_TOKEN *tk = NULL;

		filename = dd->d_name;
		len = (long) strlen( filename );

		if( (len < 4) || (strcmp( ".xml", filename +len -4 ))) {
			continue;
		} else {
			char fullpath[BUFF_MAX_SIZE];
			size_t fullsize = 0;

			PKI_CONFIG *tmp_cfg = NULL;
			char *tmp_name = NULL;

			snprintf(fullpath, BUFF_MAX_SIZE,
				"%s/%s", token_dir, filename );

			if((fullsize = strlen(token_dir) + 
				strlen( filename ) + 1) > 
						BUFF_MAX_SIZE) {
				continue;
			}
				
			if((tmp_cfg = PKI_CONFIG_load( fullpath )) ==
								NULL ) {
				continue;
			}

			if((tmp_name = PKI_CONFIG_get_value( tmp_cfg,
						"/*/name")) != NULL) {

				if((tk = PKI_TOKEN_new_null()) != NULL ) {
					if((PKI_TOKEN_init( tk, dir, tmp_name )) != PKI_ERR ) {
						PKI_STACK_TOKEN_push( list, tk);
					} else {
						PKI_TOKEN_free( tk );
					}
				}
			}

		}
	}
	closedir( dirp );

	if( url ) URL_free (url);
	if( token_dir ) PKI_Free ( token_dir );

	return ( ret );
}

/*! \brief Returns a stack of all the available Identities */
PKI_ID_INFO_STACK * PKI_list_all_id ( void ) {
	PKI_log_debug("%s:%d::Sorry, code still missing!",__FILE__,__LINE__);
	return ( NULL );
}

